#include "ppdata.h"

PPData::PPData(uchar *imageData, QSize size)
{
        m_width = size.width();
        m_height = size.height();

        m_dataMatrix = new double [m_width*m_height];
        m_dataMatrixOrig = new double [m_width*m_height];
        double *dataMatrixPtr = m_dataMatrix;
        double *dataMatrixOrigPtr = m_dataMatrixOrig;
        uchar  *imageDataPtr = imageData;

        for (int i=0; i<m_height;i++)
                for (int j=0; j<m_width;j++) {
                        *dataMatrixPtr++ = *imageDataPtr;
                        *dataMatrixOrigPtr++ = *imageDataPtr;
                        imageDataPtr += 4;

                }

        m_horProjection = new double[m_width];
        m_filteredProjection = new double[m_width];
        m_emittance = new double[MAX_PEAKS*MAX_ANGLES];
        m_peaks = new int[20];
        m_manualPeak = new int[MAX_PEAKS];
        for (int i=0; i<MAX_PEAKS; i++)
                m_manualPeak[i] = 0;
}

PPData::~PPData()
{
        delete m_dataMatrix;
        delete m_dataMatrixOrig;
        delete m_horProjection;
        delete m_filteredProjection;
        delete m_peaks;
        delete m_manualPeak;
}


/*!
    \fn int * PPData:: calcHorProjection(QSize size)
 */
double *  PPData::calcHorProjection()
{
        int i,j;

        double *projPtr = m_horProjection;
        double *pixelPtr = m_dataMatrix;
        if (m_dataMatrix == 0)
                qDebug() <<"No data" << endl;
        /*
         * first clear out the array of projection data
         */
        for (i=0; i< m_width; i++)
                *projPtr++ = 0;
        /*
         * the calculate the projection
         */

        for (i=0; i< m_height; i++) {
                projPtr = m_horProjection;
                for (j=0; j < m_width; j++) {
                        *projPtr++ += *pixelPtr++;
                }
        }

        if (m_subtractOffset) {
                double mean = average(m_horProjection,NOISE_WIDTH);
                double deviation = sigma(m_horProjection,NOISE_WIDTH);
                for (int i=0; i<m_width; i++)
                        if (m_horProjection[i] <= 0.0)
                            continue;
                        else
                            m_horProjection[i] -= (mean - deviation);
        }

        return m_horProjection;
}

/*!
    \fn int PPData:: getWidth()
 */
int PPData::getWidth()
{
        return m_width;
}

/*!
    \fn PPData::average(double *inData, int inSize)
 */
double PPData::average(double *inData, int inSize)
{
        double sum = 0;
        int count;
        for (count=0;count<m_width;count ++)
            if (inData[count] != 0.0)
                break;
//        qDebug() << "count " << count << endl;
        for (int i=count; i<count+inSize;i++) {
                sum += inData[i];
 //               qDebug() << i << " " << inData[i];
            }
        return sum/inSize;
}

/*!
    \fn PPData::sigma(double *inData, int inSize)
 */
double PPData::sigma(double *inData, int inSize)
{
        double mean = average(inData,inSize);
        double tmp,sum = 0;
        int count;
        for (count=0;count<m_width;count ++)
            if (inData[count] != 0.0)
                break;

        for (int i=count; i < count+inSize; i++) {
                tmp = inData[i];
                sum += (tmp-mean)*(tmp-mean);
        }
        sum = sum/inSize;
        return sqrt(sum);
}

/*!
    \fn PPData::maximum(double *inData, int inSize)
 */
double PPData::maximum(double *inData, int inSize)
{
        double max = 0;
        for (int i= 0; i<inSize;i++)
                if (inData[i] > max)
                        max = inData[i];
        return max;
}
/*!
    \fn PPData::getHeight()
 */
int PPData::getHeight()
{

//       printf("Data Size: %d\n",dataSize);
        return m_height;
}

/*!
    \fn double *PPData::lowPass(double *inData, int dataSize)
 */
double * PPData::lowPass(double *inData, int dataSize)
{
        const double A_COEFF = 0.05;
        const double B_COEFF = 0.1;
        const double C_COEFF = 0.2;
        const double D_COEFF = 0.3;

        m_filteredProjection[0] = inData[0];
        m_filteredProjection[1] = inData[1];
        m_filteredProjection[2] = inData[2];
        int i;
        for (i=3;i<dataSize-3;i++) {
                m_filteredProjection[i] = inData[i-3]*A_COEFF + inData[i-2]*B_COEFF + inData[i-1]*C_COEFF +
                                          inData[i]*D_COEFF +
                                          inData[i+1]*C_COEFF + inData[i+2]*B_COEFF + inData[i+3]*A_COEFF;
        }
//        printf("i: %d\n",i);
        m_filteredProjection[i] = inData[i];
        m_filteredProjection[i+1] = inData[i+1];
        m_filteredProjection[i+2] = inData[i+2];
        return m_filteredProjection;
}


/*!
    \fn PPData::findPeaks(double *inData,int inDataSize,int maxPeaks)
 */
int  PPData::findPeaks(double *inData,int inDataSize, int maxPeaks)
{
        int i,j;
        bool found;
        double threshold;
        double meanValue = average(inData,NOISE_WIDTH);
        double stdDev = sigma(inData,NOISE_WIDTH);
        double max = maximum(inData,inDataSize);
        threshold = meanValue + 4*stdDev;
        if (max/10.0 > threshold)
                threshold = max/10.0;
        m_noOfPeaks = 0;

        for (i=5; i<inDataSize-5; i++) {
                found = true;
                for (j=-5; j<5; j++)
                        if (inData[i+j] > inData[i])
                                found = false;

                if (found) {
                        /*
                         * peak must be at least 5 * the average noise
                         */
                        if (inData[i] > threshold) {
//                                printf("peak: %d\n",i);
                                m_peaks[m_noOfPeaks] = i;
                                m_noOfPeaks++;
                                if (m_noOfPeaks > maxPeaks)
                                        return -1;
                        }
                }
        }
//        qDebug() << "No of peaks found: " << m_noOfPeaks << endl;
        for (i=m_noOfPeaks; i<MAX_PEAKS;i++)
            m_peaks[i]=0;
        return m_noOfPeaks;
}


/*!
    \fn PPData::fillEmittance(bool filtered)
 */
void PPData::fillEmittance(bool filtered, bool autoFindPeaks)
{
        int holePos,distance,halfDistance,distPeakToHole,i,j,k;
        int offset = (int)(3.5*HOLE_DISTANCE*m_scalingFactor);

//        qDebug() << "Offset: " << offset << endl;
        int *emitPeaks;
        if (autoFindPeaks)
                emitPeaks = m_peaks;
        else {
                emitPeaks = m_manualPeak;
                m_noOfPeaks = MAX_PEAKS;
                sortPeaks(emitPeaks,MAX_PEAKS);
        }
//        for (int i=0;i<MAX_PEAKS;i++)
//                qDebug() << "emit peaks " << emitPeaks[i] << endl;
//        qDebug() << "m_midPoint " << m_centerPoint.x() << "scaling factor: " << m_scalingFactor << endl;

        for (i=0; i < m_noOfPeaks; i++) {
                k=0;
                if (i == 0)
                        distance = emitPeaks[1] - emitPeaks[0];
                else
                        distance = emitPeaks[i] - emitPeaks[i-1];

                halfDistance = distance/2;                              // distance between 2 peaks
                holePos = (m_centerPoint.x()-offset) +(int)((double)i*HOLE_DISTANCE*m_scalingFactor);
//                qDebug() << i << " holePos " << holePos << endl;
                distPeakToHole = holePos - emitPeaks[i];
//               qDebug() <<"Hole Position " << holePos <<  "emitPeaks[" <<i<<"]" << emitPeaks[i] << endl;
//               qDebug() <<"i " << i << " peak: " << emitPeaks[i] << "limits: " << ", " << emitPeaks[i]-halfDistance << ", " <<
//                       emitPeaks[i] + halfDistance << endl;
//               qDebug() << "limits emit: " << i << ", " << MAX_ANGLES/2+center-halfDistance << ". " << MAX_ANGLES/2+center+halfDistance << endl;

                for (j=0; j< MAX_ANGLES; j++)
                        m_emittance[i*MAX_ANGLES+j] = 0;

                for (j = emitPeaks[i]-halfDistance,k=halfDistance-distPeakToHole+ANGLE_OFFSET;
                     j < (emitPeaks[i]+halfDistance); j++,k++) {
                        if ((k<0) || (k>MAX_ANGLES)) {
                                qDebug() <<"Illegal k: " << k << endl;
                                continue;
                            }
                        else {
                            if (filtered)
                                m_emittance[i*MAX_ANGLES+k] = m_filteredProjection[j];
                            else
                                m_emittance[i*MAX_ANGLES+k] = m_horProjection[j];
                        }
                 }

        }

//       qDebug() << "fill emittance exit" << endl;
}

/*!
    \fn PPData::getEmittanceData()
 */
double * PPData::getEmittanceData()
{
        return m_emittance;
}

/*!
    \fn PPData::getDataMatrix()
 */
double * PPData::getDataMatrix()
{
        return m_dataMatrix;
}
/*!
    \fn PPData::getPeaks()
 */
int * PPData::getPeaks()
{
        return m_peaks;
}

void PPData::sortPeaks(int *peaks, int size)
{
        bool swap;
        do {
            swap = false;
            for (int i=0; i<size-1; i++)
                        if (peaks[i] > peaks[i+1]) {
                                int tmp = peaks[i];
                                peaks[i] = peaks[i+1];
                                peaks[i+1] = tmp;
                                swap=true;
                        }
        } while (swap);
        return;
}

/*!
    \fn PPData::peakSanityCheck()
 */
bool PPData::peakSanityCheck(int *peaks, int size)
{
//        qDebug() << "size: " << size << endl;
//        for (int i=0; i<MAX_PEAKS;i++)
//            qDebug() << "peak(" << i <<") " << peaks[i] << endl;

        int distance[size-1];
        int diffDistance[size-2];
        for (int i=0; i<size-1; i++) {
                distance[i]=peaks[i+1] - peaks[i];
// there should never be two identical peaks
                if (distance[i] == 0) {
                        qDebug() << " Two identical peaks! " << endl;
			return false;
                    }
                }
        for (int i=0; i<size-2;i++)
            diffDistance[i] = distance[i+1] - distance[i];
        int maxDiff = 0;
        for (int i=0; i<size-2; i++)
                if (diffDistance[i] > maxDiff)
                        maxDiff = diffDistance[i];
        if (maxDiff > MAX_DIFF) {
                qDebug() << " difference too big: " << maxDiff << endl;
                return false;
            }
        else
                return true;
}

bool PPData::peaksAllZero(int *peaks, int size)
{
        bool allZero = true;
        for (int i=0; i<size; i++)
                if (peaks[i] > 0.0)
                        allZero=false;
        return allZero;
}

int PPData::getNoOfPeaks()
{
        return m_noOfPeaks;
}

void PPData::selectAoi(int aoiX1, int aoiY1, int aoiX2, int aoiY2)
{
    int yLow,yHigh;
    int i,j;
    aoiY1 -=10;
    aoiY2 -=10;
//    qDebug() << " aoi: " << aoiX1 << " " << aoiY1 <<" " << aoiX2 << " " << aoiY2<< endl;
    if (aoiX2 < aoiX1) {
        xLow = aoiX2;
        xHigh = aoiX1;
    }
    else {
        xLow = aoiX1;
        xHigh = aoiX2;
   }
   if (aoiY2 < aoiY1) {
        yLow = aoiY2;
        yHigh = aoiY1;
    }
    else {
        yLow = aoiY1;
        yHigh = aoiY2;
    }
//    qDebug() << "x1,y1,x2,y2 " << xLow << " " << yLow << " " << xHigh << " " << yHigh << endl;
//    qDebug() << "width,height" << m_width << " " << m_height << endl;

    for (i=0; i<m_height;i++)
        for (j=0; j<m_width;j++) 
            if ((i < yLow) || (i > yHigh) || (j<xLow) || (j> xHigh))
                m_dataMatrix[i*m_width+j] = 0;
            else {
                m_dataMatrix[i*m_width+j] = m_dataMatrixOrig[i*m_width+j];

            }
}

void PPData::resetAoi()
{
    for (int i=0;i<m_width*m_height;i++)
        m_dataMatrix[i] = m_dataMatrixOrig[i];    
}
